/**********************************************************************
 *<
	FILE: ACUBIC.CPP

	DESCRIPTION: Automatic cubic map.

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mtlhdr.h"
#include "mtlres.h"
#include <bmmlib.h>
#include "render.h"
#include "stdmat.h"

//#define DBG
extern HINSTANCE hInstance;

static Class_ID acubicClassID(ACUBIC_CLASS_ID,0);

#define PB_SIZE	    0
#define PB_BLUR	    1
#define PB_BLUROFF  2

static TCHAR *getCubeName(int i)  {
	switch(i) {
		case 0: return GetString(IDS_DS_UP);  
		case 1: return GetString(IDS_DS_DOWN);  
		case 2: return GetString(IDS_DS_LEFT);  
		case 3: return GetString(IDS_DS_RIGHT);  
		case 4: return GetString(IDS_DS_TOP);  
		default: return GetString(IDS_DS_BOTTOM);  
		}
	}


Matrix3 TMForView( int i);

//---------------------------------------------------
class EnvSampler {
	public:
		virtual	BMM_Color_64 sample(Point3 n)=0;
	};

void BuildSphereMap(Bitmap *bm, EnvSampler& env) {
	FLOAT dv,du,u,v,phi,cphi,theta;;
	int iu,iv;
	Point3 n;
	int w = bm->Width();
	int h = bm->Height();
	PixelBuf l64(w);
	BMM_Color_64 *pb=l64.Ptr();

	dv = 1.0f/(float)h;
	du = 1.0f/(float)w;
	for (iv = 0,v=0.0f; iv<h; iv++,v+=dv) {
		phi = PI*(.5f - v);
		cphi = (float)cos(phi);
		n.z =  (float)sin(phi);
		for (iu=0,u=0.0f; iu<w; iu++,u+=du) {
			/* compute normal in direction for iv,iu */
			theta = TWOPI*(u-.5f);
			//if (P.reflip) theta = -theta;
			n.y = (float)sin(theta)*cphi;
			n.x = (float)cos(theta)*cphi;
			pb[iu] = env.sample(n);
			}
		bm->PutPixels(0,iv, w, pb);
		}

#ifdef DBG
	bm->Display(_T("Cubic Test"), BMM_UR);
	MessageBox(NULL, _T("The spherical map"), _T(" Auto Cubic Test"), MB_OK|MB_ICONEXCLAMATION);
	bm->UnDisplay();
#endif
	}

//---------------------------------------------------

inline float FMax(float a, float b) { return a>b?a:b; }

void ComputeSphereCoords(Point3 n, Point2& t, Point2& d) {
	t.x = 0.5f + (float)atan2(n.y,n.x)/TWOPI;
	t.y = 0.5f - (float)asin(n.z)/PI;
	float dd = FMax((float)fabs(n.x),(float)fabs(n.y))/(1.0f-n.z*n.z+.1f)/TWOPI;
	d.x = d.y = dd;
	}


//---------------------------------------------------


class ACubic;

class ACubicDlg: public ParamDlg {
	public:
		HWND hwmedit;	 	// window handle of the materials editor dialog
		IMtlParams *ip;
		ACubic *theTex;	// current ACubic being edited.
		HWND hPanel; 		// Rollup pane
		ISpinnerControl *sizeSpin,*blurSpin,*nthSpin,*blurOffSpin;
		TimeValue curTime; 
		int isActive;
		BOOL valid;

		//-----------------------------
		ACubicDlg(HWND hwMtlEdit, IMtlParams *imp, ACubic *m); 
		~ACubicDlg();
		BOOL PanelProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam );
		void LoadDialog(BOOL draw);  // stuff params into dialog
		void ReloadDialog();
		void UpdateMtlDisplay() { ip->MtlChanged(); }
		void ActivateDlg(BOOL onOff);
		void Invalidate() { valid = FALSE;	InvalidateRect(hPanel,NULL,0); }

		// methods inherited from ParamDlg:
		Class_ID ClassID() {return acubicClassID;  }
		void SetThing(ReferenceTarget *m);
		ReferenceTarget* GetThing() { return (ReferenceTarget *)theTex; }
		void DeleteThis() { delete this;  }	
		void SetTime(TimeValue t);
	};


//--------------------------------------------------------------
// CubicMap 
//--------------------------------------------------------------
#define VIEW_UP 0
#define VIEW_DN	1
#define VIEW_LF	2
#define VIEW_RT 3
#define VIEW_FR	4
#define VIEW_BK	5

class CubicMap {
	public:
		Bitmap *bitmap[6];
		Bitmap *blurmap;
		int size;
		CubicMap *next;
		int nodeID;
	// Methods
		CubicMap();
		~CubicMap();
		void FreeMaps();
		void FreeCubicMaps();
		int AllocMaps(int sz);
		BMM_Color_64 Sample(Point3 dir);
		int BuildBlurMap();
	};	  

CubicMap::CubicMap() { 
	for (int i=0; i<6; i++) 
		bitmap[i] = NULL; 
	blurmap = NULL;
	next = NULL;
	size = 0; 
	}

CubicMap::~CubicMap() { 
	FreeMaps();
	}

void CubicMap::FreeCubicMaps() {
	for (int i=0; i<6; i++) {  
		if (bitmap[i]) {
			delete bitmap[i]; 
			bitmap[i] = NULL;
			}
		}
	}

void CubicMap::FreeMaps() {
	FreeCubicMaps();
	if (blurmap) {
		delete blurmap;
		blurmap = NULL;
		}
	}

int CubicMap::AllocMaps(int sz) {
	if (bitmap[0]&&size==size)
		return 1;
	FreeMaps();
	BitmapInfo bi;
	bi.SetName(_T(""));
	bi.SetWidth(sz);
	bi.SetHeight(sz);
	bi.SetType(BMM_TRUE_32);
	size = sz;
	for (int i=0; i<6; i++) {
		if (NULL == (bitmap[i] = TheManager->Create(&bi)))
			return 0;
		}
	return 1;
	}


class CubeSampler: public EnvSampler {
	CubicMap *cm;
	public:
		CubeSampler(CubicMap *c) { cm = c; }
		BMM_Color_64 sample(Point3 n){ return cm->Sample(n);	}
	};

int CubicMap::BuildBlurMap() {
	if (bitmap[0]==NULL)
		return 1;
	if (!(blurmap&&blurmap->Width()==4*size && blurmap->Height()==2*size)) {
		BitmapInfo bi;
		bi.SetName(_T(""));
		bi.SetWidth(4*size);
		bi.SetHeight(2*size);
		bi.SetType(BMM_TRUE_32);
		if (blurmap) delete blurmap;
		blurmap = TheManager->Create(&bi);
		}
	if (blurmap) {
		CubeSampler cs(this);
		BuildSphereMap(blurmap, cs);
		FreeCubicMaps(); // TBD: Free cubic map ??
		blurmap->SetFilter(BMM_FILTER_PYRAMID);  
		BitmapFilter *filt = blurmap->Filter();
		if (filt)
			filt->MakeDirty();  // so filter gets recomputed for each frame
		}
	return 1;
	}

//--------------------------------------------------------------
// ACubic: 
//--------------------------------------------------------------

class ACubic: public StdCubic { 
	friend class ACubicPostLoad;
	friend class ACubicDlg;
	int size;
	int nth;
	float blur;
	float blurOff;
	BOOL applyBlur;
	BOOL do_nth;
	BOOL useEnvMap;
	IParamBlock *pblock;   // ref #1
	CubicMap *maps;
	TimeValue mapTime;   // when the ACubic mas were last made
	Interval ivalid;
	int rollScroll;
	ACubicDlg *paramDlg;
	public:
		ACubic();
		ParamDlg* CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp);
		void Update(TimeValue t, Interval& valid);
		void Reset();
		Interval Validity(TimeValue t) { Interval v; Update(t,v); return ivalid; }

		// From StdCubic
		void SetSize(int n, TimeValue t);
		void SetDoNth(BOOL onoff) { int do_nth = onoff; }
		void SetNth(int n) { nth = n; }
		void SetApplyBlur(BOOL onoff) { applyBlur = onoff; }
		void SetBlur(float b, TimeValue t);
		void SetBlurOffset(float b, TimeValue t);
		int GetSize(TimeValue t){ return pblock->GetInt(PB_SIZE,t); }
		BOOL GetDoNth() { return do_nth; }
		int GetNth() { return nth; }
		BOOL GetApplyBlur() { return applyBlur; }
		float GetBlur(TimeValue t) { return pblock->GetFloat(PB_BLUR,t); }
		float GetBlurOffset(TimeValue t){ return pblock->GetFloat(PB_BLUROFF,t); }

		void NotifyChanged();
	
		// Evaluate the color of map for the context.
		RGBA EvalColor(ShadeContext& sc);

		// optimized evaluation for monochrome use
		float EvalMono(ShadeContext& sc);

		// For Bump mapping, need a perturbation to apply to a normal.
		// Leave it up to the Texmap to determine how to do this.
		Point3 EvalNormalPerturb(ShadeContext& sc);

		ULONG LocalRequirements(int subMtlNum) { return MTLREQ_AUTOREFLECT;	}

		int BuildMaps(TimeValue t, RenderMapsContext &rmc);
		int DoThisFrame(TimeValue t, BOOL fieldRender);
		CubicMap *FindMap(int nodeNum);
		void FreeMaps();
		int RenderBegin(TimeValue t, ULONG flags) { mapTime = -999999999; return 1;}
		int RenderEnd(TimeValue t) { FreeMaps(); return 1; }

		Class_ID ClassID() {	return acubicClassID; }
		SClass_ID SuperClassID() { return TEXMAP_CLASS_ID; }
		void GetClassName(TSTR& s) { s= GetString(IDS_DS_ACUBIC_NAME); }  
		void DeleteThis() { delete this; }	

		int NumSubs() {return 1; }  
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return subNum; }

		// From ref
 		int NumRefs() { return 1; }
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		RefTargetHandle Clone(RemapDir &remap = NoRemap());
		RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message );

		// IO
		IOResult Save(ISave *isave);
		IOResult Load(ILoad *iload);

	};

class ACubicClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading) { 	return new ACubic; }
	const TCHAR *	ClassName() { return GetString(IDS_DS_ACUBIC_NAME); }
	SClass_ID		SuperClassID() { return TEXMAP_CLASS_ID; }
	Class_ID 		ClassID() { return acubicClassID; }
	const TCHAR* 	Category() { return TEXMAP_CAT_2D;  }
	};

static ACubicClassDesc acubicCD;

ClassDesc* GetACubicDesc() { return &acubicCD;  }

static BOOL CALLBACK  PanelDlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	ACubicDlg *theDlg;
	if (msg==WM_INITDIALOG) {
		theDlg = (ACubicDlg*)lParam;
		theDlg->hPanel = hwndDlg;
		SetWindowLong(hwndDlg, GWL_USERDATA,lParam);
		}
	else {
	    if ( (theDlg = (ACubicDlg *)GetWindowLong(hwndDlg, GWL_USERDATA) ) == NULL )
			return FALSE; 
		}
	theDlg->isActive = 1;
	int	res = theDlg->PanelProc(hwndDlg,msg,wParam,lParam);
	theDlg->isActive = 0;
	return res;
	}

ACubicDlg::ACubicDlg(HWND hwMtlEdit, IMtlParams *imp, ACubic *m) { 
	hwmedit = hwMtlEdit;
	ip = imp;
	hPanel = NULL;
	theTex = m; 
	isActive = 0;
	valid = FALSE;
	hPanel = ip->AddRollupPage( 
		hInstance,
		MAKEINTRESOURCE(IDD_AUTO_CUBIC),
		PanelDlgProc, 
		GetString(IDS_DS_ACUBIC_PARAMS), 
		(LPARAM)this );		
	curTime = imp->GetTime();
	}

void ACubicDlg::ReloadDialog() {
	Interval valid;
	theTex->Update(curTime, valid);
	LoadDialog(FALSE);
	}

void ACubicDlg::SetTime(TimeValue t) {
	Interval valid;
	if (t!=curTime) {
		curTime = t;
		theTex->Update(curTime, valid);
		LoadDialog(FALSE);
		InvalidateRect(hPanel,NULL,0);
		}
	}

ACubicDlg::~ACubicDlg() {
	theTex->paramDlg = NULL;
	ReleaseISpinner(sizeSpin);
	ReleaseISpinner(blurSpin);
	ReleaseISpinner(blurOffSpin);
	ReleaseISpinner(nthSpin);
	SetWindowLong(hPanel, GWL_USERDATA, NULL);
	hPanel =  NULL;
	}


BOOL ACubicDlg::PanelProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam ) {
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);
    switch (msg)    {
		case WM_INITDIALOG:
			{
			sizeSpin = SetupIntSpinner(hwndDlg, IDC_ACUBE_SIZE_SPIN, IDC_ACUBE_SIZE_EDIT,1,5000,100);
			blurSpin = SetupFloatSpinner(hwndDlg, IDC_ACUBE_BLUR_SPIN, IDC_ACUBE_BLUR_EDIT,0.0f,100.0f,1.0f,.01f);
			blurOffSpin = SetupFloatSpinner(hwndDlg, IDC_ACUBE_BLUROFF_SPIN, IDC_ACUBE_BLUOFF_EDIT,0.0f,1.0f,0.0f,.001f);
			nthSpin = SetupIntSpinner(hwndDlg, IDC_ACUBE_NTH_SPIN, IDC_ACUBE_NTH_EDIT,1,1000, 1);
			CheckRadioButton( hwndDlg, IDC_FIRST_ONLY, IDC_EVERY_NTH, IDC_FIRST_ONLY+theTex->do_nth);
			SetCheckBox(hwndDlg, IDC_ACUBE_BLUR, theTex->applyBlur);
			SetCheckBox(hwndDlg, IDC_USE_ENVMAP, theTex->useEnvMap);
			return TRUE;
			}
			break;
		case WM_COMMAND:  
		    switch (id) {
				case IDC_FIRST_ONLY:
					theTex->do_nth = FALSE;
					break;
				case IDC_EVERY_NTH:
					theTex->do_nth = TRUE;
					break;
				case IDC_ACUBE_BLUR:
					theTex->applyBlur = GetCheckBox(hwndDlg, id);			
					break;
				case IDC_USE_ENVMAP:
					theTex->useEnvMap = GetCheckBox(hwndDlg, id);			
					break;
				}
			break;
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			ip->RollupMouseMessage(hwndDlg,msg,wParam,lParam);
			return FALSE;
		case WM_PAINT: 	
			if (!valid) {
				valid = TRUE;
				ReloadDialog();
				}
			break;
		case WM_CLOSE: 	break;       
		case WM_DESTROY:		break;
		case CC_SPINNER_CHANGE: 
			if (!theHold.Holding()) theHold.Begin();
			switch (id) {
				case IDC_ACUBE_SIZE_SPIN: 
					theTex->SetSize(sizeSpin->GetIVal(),curTime); 	
					break;
				case IDC_ACUBE_BLUR_SPIN: 
					theTex->SetBlur(blurSpin->GetFVal(),curTime); 	
					break;
				case IDC_ACUBE_BLUROFF_SPIN: 
					theTex->SetBlurOffset(blurOffSpin->GetFVal(),curTime); 	
					break;
				case IDC_ACUBE_NTH_SPIN: 
					theTex->nth = nthSpin->GetIVal(); 	
					break;
				}
			break;
		case CC_SPINNER_BUTTONDOWN:
			theHold.Begin();
			break;
		case WM_CUSTEDIT_ENTER:
		case CC_SPINNER_BUTTONUP: 
			if (HIWORD(wParam) || msg==WM_CUSTEDIT_ENTER) theHold.Accept(GetString(IDS_DS_PARAMCHG));
			else theHold.Cancel();
			theTex->NotifyChanged();
		    UpdateMtlDisplay();
			break;
    	}
	return FALSE;
	}
 

void ACubicDlg::LoadDialog(BOOL draw) {
	if (theTex) {
		Interval valid;
		theTex->Update(curTime,valid);
		sizeSpin->SetValue(theTex->size,FALSE);
		blurSpin->SetValue(theTex->blur,FALSE);
		blurOffSpin->SetValue(theTex->blurOff,FALSE);
		nthSpin->SetValue(theTex->nth,FALSE);
		CheckRadioButton( hPanel, IDC_FIRST_ONLY, IDC_EVERY_NTH, IDC_FIRST_ONLY+theTex->do_nth);
		SetCheckBox(hPanel, IDC_ACUBE_BLUR, theTex->applyBlur);
		}
	}

void ACubicDlg::SetThing(ReferenceTarget *m) {
	assert (m->ClassID()==acubicClassID);
	assert (m->SuperClassID()==TEXMAP_CLASS_ID);
	if (theTex) theTex->paramDlg = NULL;
	theTex = (ACubic *)m;
	if (theTex) theTex->paramDlg = this;
	LoadDialog(TRUE);
	}

void ACubicDlg::ActivateDlg(BOOL onOff) {
//	for (int i=0; i<NCOLS; i++)
//		cs[i]->Activate(onOff);
	}

//-----------------------------------------------------------------------------
//  ACubic
//-----------------------------------------------------------------------------


#define NPARAMS 3
#define ACUBIC_VERSION 2


// Version 1 desc
static ParamBlockDescID pbdesc1[] = {
	{ TYPE_INT, NULL, TRUE,0 }, 	// size
	{ TYPE_FLOAT, NULL, TRUE,0 } 	// blur
	};

// Version 2 desc
static ParamBlockDescID pbdesc2[] = {
	{ TYPE_INT, NULL, TRUE,0 }, 	// size
	{ TYPE_FLOAT, NULL, TRUE,0 }, 	// blur
	{ TYPE_FLOAT, NULL, TRUE,0 } 	// blurOffs
	};

void ACubic::Reset() {
	ReplaceReference( 0, CreateParameterBlock( pbdesc2, NPARAMS, ACUBIC_VERSION) );	
	ivalid.SetEmpty();
	SetSize(100, TimeValue(0));
	SetBlur(1.0f, TimeValue(0));
	SetBlurOffset(.0f, TimeValue(0));
	}

void ACubic::NotifyChanged() {
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
	}

ACubic::ACubic() {
	paramDlg = NULL;
	pblock = NULL;
	nth = 1;
	do_nth = TRUE;
	applyBlur = TRUE;
	useEnvMap = TRUE;
	maps = NULL;
	Reset();
	rollScroll=0;
	}

RefTargetHandle ACubic::Clone(RemapDir &remap) {
	ACubic *mnew = new ACubic();
	*((MtlBase*)mnew) = *((MtlBase*)this);  // copy superclass stuff
	mnew->ReplaceReference(0,remap.CloneRef(pblock));
	mnew->do_nth = do_nth;
	mnew->applyBlur = applyBlur;
	mnew->ivalid.SetEmpty();	
	return (RefTargetHandle)mnew;
	}

ParamDlg* ACubic::CreateParamDlg(HWND hwMtlEdit, IMtlParams *imp) {
	ACubicDlg *dm = new ACubicDlg(hwMtlEdit, imp, this);
	dm->LoadDialog(TRUE);	
	paramDlg = dm;
	return dm;	
	}

void ACubic::Update(TimeValue t, Interval& valid) {		
	if (!ivalid.InInterval(t)) {
		ivalid.SetInfinite();
		pblock->GetValue( PB_SIZE, t, size, ivalid );
		pblock->GetValue( PB_BLUR, t, blur, ivalid );
		pblock->GetValue( PB_BLUROFF, t, blurOff, ivalid );
		}
	valid &= ivalid;
	}

void ACubic::FreeMaps() {
	CubicMap *cm,*nxtcm;
	for (cm = maps; cm!=NULL; cm = nxtcm) {
		nxtcm = cm->next;
	   	delete cm;		
		}
	maps = NULL;
	}


void ACubic::SetSize(int s, TimeValue t) { 
	size = s; 
	pblock->SetValue( PB_SIZE, t, s);
	}

void ACubic::SetBlur(float f, TimeValue t) { 
	blur = f; 
	pblock->SetValue( PB_BLUR, t, f);
	}

void ACubic::SetBlurOffset(float f, TimeValue t) { 
	blurOff = f; 
	pblock->SetValue( PB_BLUROFF, t, f);
	}

RefTargetHandle ACubic::GetReference(int i) {
	switch(i) {
		case 0:	return pblock ;
		default: return NULL;
		}
	}

void ACubic::SetReference(int i, RefTargetHandle rtarg) {
	switch(i) {
		case 0:	pblock = (IParamBlock *)rtarg; break;
		}
	}
	 
Animatable* ACubic::SubAnim(int i) {
	switch (i) {
		case 0: return pblock;
		default: return NULL;
		}
	}

TSTR ACubic::SubAnimName(int i) {
	switch (i) {
		case 0: return TSTR(GetString(IDS_DS_PARAMETERS));
		default: return TSTR("");		
		}
	}

static int nameID[] = {IDS_DS_MAPSIZE, IDS_DS_BLUR, IDS_DS_BLUROFFS };

RefResult ACubic::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
   PartID& partID, RefMessage message ) {
	switch (message) {
		case REFMSG_CHANGE:
			ivalid.SetEmpty();
			if (paramDlg&&!paramDlg->isActive) 
				paramDlg->Invalidate();
			break;

		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_SIZE: 
				case PB_BLUR: 
				case PB_BLUROFF: 
					gpd->dim = defaultDim; break;
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			gpn->name = TSTR(GetString(nameID[gpn->index]));
			return REF_STOP; 
			}
		}
	return(REF_SUCCEED);
	}


inline Point2 CompUV(float x, float y, float z) {
	return Point2( 0.5f*(x/z+1.0f), 0.5f*(y/z+1.0f));
	}

inline AColor AColorFrom64(BMM_Color_64 pix) {
	return AColor((float)pix.r/65536.0f,(float)pix.g/65536.0f,(float)pix.b/65536.0f,(float)pix.a/65536.0f);
	}


// the 
Matrix3 TMForView( int i) {
	Matrix3 m;
	m.IdentityMatrix();
	// We start out looking into the Z-axis (top view) 
	switch (i) {
		case VIEW_UP:    // looking UP
			m.RotateX(-PI);	
			break;
		case VIEW_DN:    // looking down (top view)
			break;
		case VIEW_LF: 	 // looking to left
			m.RotateX(-.5f*PI);	
			m.RotateY(-.5f*PI);
			break;
		case VIEW_RT:     // looking to right
			m.RotateX(-.5f*PI);	
			m.RotateY(+.5f*PI);
			break;
		case VIEW_FR:       // looking to front (back view)
			m.RotateX(-.5f*PI);	
			m.RotateY(PI);
			break;
		case VIEW_BK:         // looking to back
			m.RotateX(-.5f*PI);	
			break;
		}
	return m;
	}

int ACubic::DoThisFrame(TimeValue t, BOOL fieldRender) {
	if (t==mapTime) return 1;  // must be another object with this map on the same frame.
	if (!do_nth) return 0;  // only do it once.
	if (nth==1) return 1;   // need every one
	TimeValue del = abs(t - mapTime);
	if (fieldRender) del*=2;
	return  (del>=SecToTicks(nth))?1:0;
	}

CubicMap *ACubic::FindMap(int nodeNum) {
	CubicMap *cm;
	for (cm=maps; cm!=NULL; cm = cm->next)
		if (cm->nodeID==nodeNum) return cm;
	return NULL;
	}

int ACubic::BuildMaps(TimeValue t, RenderMapsContext &rmc) {
	SubRendParams srp;
	rmc.GetSubRendParams(srp);
	CubicMap *cm = FindMap(rmc.NodeRenderID());
	if (cm&&!DoThisFrame(t,srp.fieldRender))
		return 1;
	mapTime = t;
	INode *node = rmc.GetNode();
	Matrix3 tm = node->GetNodeTM(t);
	Point3 pos = tm.GetRow(3); //nodes world position
	ViewParams vp;
	vp.projType = PROJ_PERSPECTIVE;
	vp.hither = .001f;
	vp.yon = 1.0e30f;
	vp.fov = PI/2.0f;

	srp.devWidth = size;
	srp.devHeight = size;
	srp.devAspect = 1.0f;
	srp.xmin = srp.ymin = 0;
	srp.xmax = srp.ymax = size;
	srp.xorg =srp.yorg = 0;
	srp.fieldRender = 0;
	srp.rendType = RENDTYPE_NORMAL;
	srp.doEnvMap = useEnvMap;
	srp.doingMirror = FALSE;
	if (cm==NULL) {
		cm = new CubicMap;
		cm->nodeID = rmc.NodeRenderID();
		cm->next = maps;
		maps = cm;
		}
	if (cm->AllocMaps(size)) {
		for (int i=0; i<6; i++) {	
			tm = TMForView(i);
			tm.PreTranslate(-pos); 
			vp.affineTM = tm;
			if (!rmc.Render(cm->bitmap[i], vp, srp, 0)) {
				cm->FreeMaps();
				return 0;
				}
#ifdef DBG
			cm->bitmap[i]->Display(getCubeName(i), BMM_UR);
			MessageBox(NULL, getCubeName(i), _T(" Auto Cubic Test"), MB_OK|MB_ICONEXCLAMATION);
			cm->bitmap[i]->UnDisplay();
#endif
			}
		}

	if (applyBlur) {
		cm->BuildBlurMap();
		}
	return 1;
	}

static BMM_Color_64 black64 = {0,0,0,0};
// The "canonical" view is looking in the -z direction,
// with x to the right, y up, i.e. VIEW_DN,
// other views are mapped into this canonical rep

BMM_Color_64 CubicMap::Sample(Point3 v) {
	float wx,wy,wz;
	Color rcol;
	Bitmap *refmap=NULL;
	Point3 rv;
	Point2 uv;
	wx = (float)fabs(v.x);  wy = (float)fabs(v.y);  wz = (float)fabs(v.z); 
	if (wx>=wy && wx>=wz) {
		if(v.x<0) {	refmap = bitmap[VIEW_LF];	uv = CompUV(-v.y, -v.z,  v.x);	}
		else	  { refmap = bitmap[VIEW_RT];   uv = CompUV( v.y, -v.z, -v.x);	}
		}
	else
	if(wy>=wx && wy>=wz) {
		if(v.y>0) { refmap = bitmap[VIEW_BK];	uv = CompUV( -v.x, -v.z, -v.y);	}
		else {  	refmap = bitmap[VIEW_FR];	uv = CompUV(  v.x, -v.z,  v.y); }
		}
	else
	if(wz>=wx && wz>=wy) {
		if(v.z<0) {	refmap = bitmap[VIEW_DN];	uv = CompUV( -v.x, -v.y,  v.z);	}
		else     {	refmap = bitmap[VIEW_UP];	uv = CompUV( -v.x,  v.y, -v.z);	}
		}

	if (refmap==NULL) return black64;
	int x = (int)(frac(uv.x)*(float)size);
	int y = (int)(frac(1.0f-uv.y)*(float)size);
	BMM_Color_64 c;
	refmap->GetPixels(x,y,1,&c);
	return c;
	}

RGBA ACubic::EvalColor(ShadeContext& sc) {
	if (gbufID) sc.SetGBufferID(gbufID);
	Point3 V = sc.VectorTo(sc.V(),REF_WORLD);
	int id = sc.NodeID();
	CubicMap *cm = FindMap(id);
	if (cm) {
		if (applyBlur&&cm->blurmap) {
			BMM_Color_64 c;
			Point2 uv, duv;
			ComputeSphereCoords(V,uv,duv);
			duv *= (blur*sc.Curve());
			duv.x += blurOff;
			duv.y += blurOff;
			cm->blurmap->GetFiltered(uv.x,uv.y,duv.x,duv.y,&c);
			return AColorFrom64(c);
			}
		else 
			return AColorFrom64(cm->Sample(V));
		}
	else 
		return Color(0,0,0);
	}

float ACubic::EvalMono(ShadeContext& sc) {
	return Intens(EvalColor(sc));
	}

Point3 ACubic::EvalNormalPerturb(ShadeContext& sc) {
	return Point3(0,0,0);
	}

#define MTL_HDR_CHUNK 0x4000
#define DONT_DO_NTH_CHUNK 0x1000
#define NTH_CHUNK 0x1001
#define DONT_APPLY_BLUR_CHUNK 0x1002
#define DONT_USE_ENV_CHUNK 0x1003

IOResult ACubic::Save(ISave *isave) { 
	IOResult res;
	ULONG nb;
	// Save common stuff
	isave->BeginChunk(MTL_HDR_CHUNK);
	res = MtlBase::Save(isave);
	if (res!=IO_OK) return res;
	isave->EndChunk();
	if (!do_nth) {
		isave->BeginChunk(DONT_DO_NTH_CHUNK);
		isave->EndChunk();
		}
	if (!applyBlur) {
		isave->BeginChunk(DONT_APPLY_BLUR_CHUNK);
		isave->EndChunk();
		}
	if (!useEnvMap) {
		isave->BeginChunk(DONT_USE_ENV_CHUNK);
		isave->EndChunk();
		}
	isave->BeginChunk(NTH_CHUNK);
	isave->Write(&nth,sizeof(nth),&nb);			
	isave->EndChunk();
	return IO_OK;
	}	

class ACubicPostLoad : public PostLoadCallback {
	public:
		ACubic *cube;
		ACubicPostLoad(ACubic *b) {cube=b;}
		void proc(ILoad *iload) {
			if (cube->pblock->GetVersion()!=ACUBIC_VERSION) {
				switch (cube->pblock->GetVersion()) {
					case 1:
						cube->ReplaceReference(0,
							UpdateParameterBlock(
								pbdesc1, 2, cube->pblock,
								pbdesc2, 3, ACUBIC_VERSION));
						iload->SetObsolete();
						break;
					default:
						assert(0);
						break;
					}
				}
			//waitPostLoad--;
			delete this;
			}
	};

IOResult ACubic::Load(ILoad *iload) { 
	ULONG nb;
	IOResult res;
	iload->RegisterPostLoadCallback(new ACubicPostLoad(this));
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case MTL_HDR_CHUNK:
				res = MtlBase::Load(iload);
				break;
			case DONT_DO_NTH_CHUNK:
				do_nth = FALSE;
				break;
			case DONT_APPLY_BLUR_CHUNK:
				applyBlur = FALSE;
				break;
			case DONT_USE_ENV_CHUNK:
				useEnvMap = FALSE;
				break;
			case NTH_CHUNK:
				iload->Read(&nth,sizeof(nth),&nb);			
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	return IO_OK;
	}
